package ru.octol1ttle.knockdowns;

import java.util.UUID;
import net.fabricmc.api.ClientModInitializer;
import net.fabricmc.fabric.api.client.event.lifecycle.v1.ClientEntityEvents;
import net.fabricmc.fabric.api.client.event.lifecycle.v1.ClientTickEvents;
import net.fabricmc.fabric.api.client.networking.v1.ClientPlayNetworking;
import net.fabricmc.fabric.api.client.rendering.v1.HudRenderCallback;
import net.fabricmc.fabric.api.event.player.UseEntityCallback;
import net.fabricmc.fabric.api.networking.v1.PacketByteBufs;
import net.minecraft.SharedConstants;
import net.minecraft.client.MinecraftClient;
import net.minecraft.client.font.TextRenderer;
import net.minecraft.network.PacketByteBuf;
import net.minecraft.util.ActionResult;
import net.minecraft.util.hit.EntityHitResult;
import net.minecraft.util.hit.HitResult;
import ru.octol1ttle.knockdowns.api.IKnockableDown;
import ru.octol1ttle.knockdowns.network.KnockdownsNetworkingConstants;

public class KnockdownsClient implements ClientModInitializer {
    private static final int REVIVAL_WAIT_TIME = 10 * SharedConstants.TICKS_PER_SECOND;
    private static IKnockableDown reviving = null;
    private static int revivalTimer = -1;

    @Override
    public void onInitializeClient() {
        registerOnReceiveKnockedDown();
        registerOnReceiveReviving();
        registerOnEntityLoad();
        registerOnEntityUse();
        registerOnStartWorldTick();
        registerOnHudRender();
    }

    private static void registerOnReceiveKnockedDown() {
        ClientPlayNetworking.registerGlobalReceiver(KnockdownsNetworkingConstants.S2C_SEND_PLAYER_KNOCKED_DOWN, (client, handler, buf, responseSender) -> {
            UUID uuid = buf.readUuid();
            boolean knockedDown = buf.readBoolean();

            client.execute(() -> {
                IKnockableDown knockableDown = (IKnockableDown) handler.getWorld().getPlayerByUuid(uuid);
                if (knockableDown != null) {
                    knockableDown.knockdowns$setKnockedDown(knockedDown);
                }
            });
        });
    }

    private static void registerOnReceiveReviving() {
        ClientPlayNetworking.registerGlobalReceiver(KnockdownsNetworkingConstants.S2C_SEND_PLAYER_REVIVING, (client, handler, buf, responseSender) -> {
            UUID uuid = buf.readUuid();
            boolean beingRevived = buf.readBoolean();

            client.execute(() -> {
                IKnockableDown knockableDown = (IKnockableDown) handler.getWorld().getPlayerByUuid(uuid);
                if (knockableDown != null) {
                    knockableDown.knockdowns$setBeingRevived(beingRevived);

                    if (client.player != null && uuid.equals(client.player.getUuid())) {
                        if (knockableDown.knockdowns$isBeingRevived()) {
                            revivalTimer = REVIVAL_WAIT_TIME;
                        } else {
                            revivalTimer = -1;
                        }
                    }

                    if (reviving != null && knockableDown.knockdowns$getUuid().equals(reviving.knockdowns$getUuid())
                            && !knockableDown.knockdowns$isBeingRevived()) {
                        reviving = null;
                        revivalTimer = -1;
                    }
                }
            });
        });
    }

    private static void registerOnEntityLoad() {
        ClientEntityEvents.ENTITY_LOAD.register((entity, world) -> {
            if (entity instanceof IKnockableDown) {
                PacketByteBuf buf = PacketByteBufs.create();
                buf.writeUuid(entity.getUuid());

                ClientPlayNetworking.send(KnockdownsNetworkingConstants.C2S_REQUEST_PLAYER_KNOCKED_DOWN, buf);
                ClientPlayNetworking.send(KnockdownsNetworkingConstants.C2S_REQUEST_PLAYER_REVIVING, buf);
            }
        });
    }

    private static void registerOnEntityUse() {
        UseEntityCallback.EVENT.register((player, world, hand, entity, hitResult) -> {
            if (!(entity instanceof IKnockableDown knockableDown) || !knockableDown.knockdowns$isKnockedDown()
                    || knockableDown.knockdowns$isBeingRevived()) {
                return ActionResult.PASS;
            }
            if (((IKnockableDown)player).knockdowns$isKnockedDown()) {
                return ActionResult.FAIL;
            }

            knockableDown.knockdowns$setBeingRevived(true);

            PacketByteBuf buf = PacketByteBufs.create();
            buf.writeUuid(entity.getUuid());
            buf.writeBoolean(true);

            ClientPlayNetworking.send(KnockdownsNetworkingConstants.C2S_SEND_PLAYER_REVIVING, buf);

            reviving = knockableDown;
            revivalTimer = REVIVAL_WAIT_TIME;

            return ActionResult.SUCCESS;
        });
    }

    private static void registerOnStartWorldTick() {
        ClientTickEvents.START_WORLD_TICK.register(world -> {
            boolean revived = false;
            revivalTimer--;
            if (revivalTimer <= 0) {
                revivalTimer = -1;
                revived = true;
            }
            
            if (reviving == null) {
                return;
            }

            HitResult crosshairTarget = MinecraftClient.getInstance().crosshairTarget;
            if (revived || crosshairTarget == null || crosshairTarget.getType() != HitResult.Type.ENTITY
                    || !((EntityHitResult) crosshairTarget).getEntity().getUuid().equals(reviving.knockdowns$getUuid())) {
                reviving.knockdowns$setBeingRevived(false);

                PacketByteBuf buf = PacketByteBufs.create();
                buf.writeUuid(reviving.knockdowns$getUuid());
                buf.writeBoolean(false);

                ClientPlayNetworking.send(KnockdownsNetworkingConstants.C2S_SEND_PLAYER_REVIVING, buf);
                if (revived) {
                    reviving.knockdowns$setKnockedDown(false);

                    PacketByteBuf revivedBuf = PacketByteBufs.create();
                    revivedBuf.writeUuid(reviving.knockdowns$getUuid());

                    ClientPlayNetworking.send(KnockdownsNetworkingConstants.C2S_SEND_PLAYER_REVIVED, revivedBuf);
                }

                reviving = null;
                revivalTimer = -1;
            }
        });
    }

    private static void registerOnHudRender() {
        HudRenderCallback.EVENT.register((drawContext, tickDelta) -> {
            if (revivalTimer == -1) {
                return;
            }

            TextRenderer renderer = MinecraftClient.getInstance().textRenderer;
            String text = String.format("%.1f", revivalTimer / (float) SharedConstants.TICKS_PER_SECOND);
            int x = (drawContext.getScaledWindowWidth() - renderer.getWidth(text)) / 2;

            drawContext.drawTextWithShadow(renderer, text, x, drawContext.getScaledWindowHeight() / 2 + 15, 0xFFFFFF);
        });
    }
}