package ru.octol1ttle.knockdowns;

import java.util.UUID;
import net.fabricmc.api.ModInitializer;
import net.fabricmc.fabric.api.entity.event.v1.ServerLivingEntityEvents;
import net.fabricmc.fabric.api.event.player.AttackBlockCallback;
import net.fabricmc.fabric.api.event.player.AttackEntityCallback;
import net.fabricmc.fabric.api.event.player.UseBlockCallback;
import net.fabricmc.fabric.api.event.player.UseItemCallback;
import net.fabricmc.fabric.api.networking.v1.PacketByteBufs;
import net.fabricmc.fabric.api.networking.v1.PlayerLookup;
import net.fabricmc.fabric.api.networking.v1.ServerPlayNetworking;
import net.minecraft.entity.player.PlayerEntity;
import net.minecraft.network.PacketByteBuf;
import net.minecraft.network.packet.s2c.play.PlaySoundS2CPacket;
import net.minecraft.registry.Registries;
import net.minecraft.registry.Registry;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.network.ServerPlayerEntity;
import net.minecraft.sound.SoundCategory;
import net.minecraft.sound.SoundEvent;
import net.minecraft.text.Text;
import net.minecraft.text.TranslatableTextContent;
import net.minecraft.util.ActionResult;
import net.minecraft.util.Hand;
import net.minecraft.util.Identifier;
import net.minecraft.util.TypedActionResult;
import net.minecraft.world.GameRules;
import ru.octol1ttle.knockdowns.api.IKnockableDown;
import ru.octol1ttle.knockdowns.network.KnockdownsNetworkingConstants;

public class Knockdowns implements ModInitializer {
    public static final String MOD_ID = "knockdowns";
    private static final Identifier KNOCKED_DOWN_ID = new Identifier("knockdowns:knocked_down");
    private static final SoundEvent KNOCKED_DOWN = SoundEvent.of(KNOCKED_DOWN_ID);

    @Override
    public void onInitialize() {
        registerSoundEvents();
        registerOnDeath();
        registerOnRequestKnockedDown();
        registerOnRequestReviving();
        registerOnReceiveReviving();
        registerOnReceiveRevived();
        registerOnPlayerInteractions();
    }

    private static void registerSoundEvents() {
        Registry.register(Registries.SOUND_EVENT, KNOCKED_DOWN_ID, KNOCKED_DOWN);
    }

    private static void registerOnDeath() {
        ServerLivingEntityEvents.ALLOW_DEATH.register((entity, damageSource, damageAmount) -> {
            if (!(entity instanceof IKnockableDown knockableDown) || knockableDown.knockdowns$isKnockedDown()) {
                return true;
            }

            ServerPlayerEntity serverPlayer = (ServerPlayerEntity) entity;
            // TODO: timer
            if (!serverPlayer.getWorld().getGameRules().getBoolean(GameRules.KEEP_INVENTORY)) {
                serverPlayer.getInventory().dropAll();
            }
            entity.setHealth(1.0f);
            entity.setInvulnerable(true);
            entity.setGlowing(true);
            entity.setAir(entity.getMaxAir());
            entity.extinguish();
            entity.setFrozenTicks(0);
            entity.setOnFire(false);
            entity.clearStatusEffects();

            knockableDown.knockdowns$setKnockedDown(true);

            PacketByteBuf buf = PacketByteBufs.create();
            buf.writeUuid(entity.getUuid());
            buf.writeBoolean(true);

            ServerPlayNetworking.send(serverPlayer, KnockdownsNetworkingConstants.S2C_SEND_PLAYER_KNOCKED_DOWN, buf);
            for (ServerPlayerEntity player : PlayerLookup.tracking(entity)) {
                ServerPlayNetworking.send(player, KnockdownsNetworkingConstants.S2C_SEND_PLAYER_KNOCKED_DOWN, buf);
            }

            for (ServerPlayerEntity player : serverPlayer.getServerWorld().getPlayers()) {
                player.networkHandler.sendPacket(new PlaySoundS2CPacket(Registries.SOUND_EVENT.getEntry(KNOCKED_DOWN), SoundCategory.PLAYERS,
                        player.getX(), player.getY(), player.getZ(), 1.0f, 1.0f, 0L));
            }

            TranslatableTextContent content = (TranslatableTextContent) entity.getDamageTracker().getDeathMessage().getContent();
            Text replaced = Text.translatable(content.getKey().replace("death.", "knockdown."), content.getArgs());
            MinecraftServer server = serverPlayer.getServer();
            if (server != null) {
                server.getPlayerManager().broadcast(replaced, false);
            }

            return false;
        });
    }

    private static void registerOnRequestKnockedDown() {
        ServerPlayNetworking.registerGlobalReceiver(KnockdownsNetworkingConstants.C2S_REQUEST_PLAYER_KNOCKED_DOWN, (server, player, handler, buf, responseSender) -> {
            UUID uuid = buf.readUuid();

            server.execute(() -> {
                IKnockableDown knockableDown = (IKnockableDown) player.getWorld().getPlayerByUuid(uuid);
                if (knockableDown == null) {
                    return;
                }

                PacketByteBuf responseBuf = PacketByteBufs.create();
                responseBuf.writeUuid(uuid);
                responseBuf.writeBoolean(knockableDown.knockdowns$isKnockedDown());

                responseSender.sendPacket(KnockdownsNetworkingConstants.S2C_SEND_PLAYER_KNOCKED_DOWN, responseBuf);
            });
        });
    }

    private static void registerOnRequestReviving() {
        ServerPlayNetworking.registerGlobalReceiver(KnockdownsNetworkingConstants.C2S_REQUEST_PLAYER_REVIVING, (server, player, handler, buf, responseSender) -> {
            UUID uuid = buf.readUuid();

            server.execute(() -> {
                IKnockableDown knockableDown = (IKnockableDown) player.getWorld().getPlayerByUuid(uuid);
                if (knockableDown == null) {
                    return;
                }

                PacketByteBuf responseBuf = PacketByteBufs.create();
                responseBuf.writeUuid(uuid);
                responseBuf.writeBoolean(knockableDown.knockdowns$isBeingRevived());

                responseSender.sendPacket(KnockdownsNetworkingConstants.S2C_SEND_PLAYER_REVIVING, responseBuf);
            });
        });
    }

    private static void registerOnReceiveReviving() {
        ServerPlayNetworking.registerGlobalReceiver(KnockdownsNetworkingConstants.C2S_SEND_PLAYER_REVIVING, (server, player, handler, buf, responseSender) -> {
            UUID uuid = buf.readUuid();
            boolean beingRevived = buf.readBoolean();

            // TODO: revival by multiple players
            server.execute(() -> {
                PlayerEntity reviving = player.getWorld().getPlayerByUuid(uuid);
                IKnockableDown knockableDown = (IKnockableDown) reviving;
                if (knockableDown == null) {
                    return;
                }

                knockableDown.knockdowns$setBeingRevived(beingRevived);

                PacketByteBuf broadcastBuf = PacketByteBufs.create();
                broadcastBuf.writeUuid(uuid);
                broadcastBuf.writeBoolean(beingRevived);

                ServerPlayNetworking.send((ServerPlayerEntity) reviving, KnockdownsNetworkingConstants.S2C_SEND_PLAYER_REVIVING, broadcastBuf);
                for (ServerPlayerEntity entity : PlayerLookup.tracking(reviving)) {
                    ServerPlayNetworking.send(entity, KnockdownsNetworkingConstants.S2C_SEND_PLAYER_REVIVING, broadcastBuf);
                }
            });
        });
    }

    private static void registerOnReceiveRevived() {
        ServerPlayNetworking.registerGlobalReceiver(KnockdownsNetworkingConstants.C2S_SEND_PLAYER_REVIVED, (server, player, handler, buf, responseSender) -> {
            UUID uuid = buf.readUuid();

            server.execute(() -> {
                PlayerEntity reviving = player.getWorld().getPlayerByUuid(uuid);
                IKnockableDown knockableDown = (IKnockableDown) reviving;
                if (knockableDown == null || !knockableDown.knockdowns$isKnockedDown()) {
                    return;
                }

                reviving.setInvulnerable(false);
                reviving.setGlowing(false);
                reviving.setHealth(reviving.getHealth() + 5.0f);

                knockableDown.knockdowns$setKnockedDown(false);

                PacketByteBuf sendBuf = PacketByteBufs.create();
                sendBuf.writeUuid(reviving.getUuid());
                sendBuf.writeBoolean(false);

                ServerPlayNetworking.send((ServerPlayerEntity) reviving, KnockdownsNetworkingConstants.S2C_SEND_PLAYER_KNOCKED_DOWN, sendBuf);
                for (ServerPlayerEntity entity : PlayerLookup.tracking(reviving)) {
                    ServerPlayNetworking.send(entity, KnockdownsNetworkingConstants.S2C_SEND_PLAYER_KNOCKED_DOWN, sendBuf);
                }
            });
        });
    }

    private static void registerOnPlayerInteractions() {
        AttackBlockCallback.EVENT.register((player, world, hand, pos, direction) -> {
            if (player instanceof IKnockableDown && ((IKnockableDown) player).knockdowns$isKnockedDown()) {
                return ActionResult.FAIL;
            }
            return ActionResult.PASS;
        });
        AttackEntityCallback.EVENT.register((player, world, hand, entity, hitResult) -> {
            if (player instanceof IKnockableDown && ((IKnockableDown) player).knockdowns$isKnockedDown()) {
                return ActionResult.FAIL;
            }
            return ActionResult.PASS;
        });
        UseItemCallback.EVENT.register((player, world, hand) -> {
            if (player instanceof IKnockableDown && ((IKnockableDown) player).knockdowns$isKnockedDown()) {
                return TypedActionResult.fail(hand == Hand.MAIN_HAND ? player.getMainHandStack() : player.getOffHandStack());
            }
            return TypedActionResult.pass(hand == Hand.MAIN_HAND ? player.getMainHandStack() : player.getOffHandStack());
        });
        UseBlockCallback.EVENT.register((player, world, hand, hitResult) -> {
            if (player instanceof IKnockableDown && ((IKnockableDown) player).knockdowns$isKnockedDown()) {
                return ActionResult.FAIL;
            }
            return ActionResult.PASS;
        });
    }
}